from typing import Generic, TypeVar, Dict

from apscheduler.triggers.base import BaseTrigger

from web_foundation.environment.workers.background.worker import BackgroundTask, TaskErrorCallback, TaskIMessage
from web_foundation.kernel import IMessage
from web_foundation.kernel.worker import GenWorker


class Service(Generic[GenWorker]):
    _worker: GenWorker

    async def run_background(self, task: BackgroundTask,
                             *args,
                             trigger: BaseTrigger = None,
                             on_error_callback: TaskErrorCallback = None,
                             add_job_kw: Dict = None, **kwargs):
        await self.emmit_event(
            TaskIMessage(task,
                         trigger,
                         on_error_callback=on_error_callback,
                         add_job_kw=add_job_kw,
                         args=args,
                         kwargs=kwargs))

    async def wait_for_response(self, msg: IMessage):
        return await self.worker.channel.produce_for_response(msg)

    @property
    def worker(self) -> GenWorker:
        return self._worker

    @worker.setter
    def worker(self, worker: GenWorker):
        self._worker = worker

    async def emmit_event(self, event: IMessage):
        await self.worker.channel.produce(event)


# async def microtask(self, task: Coroutine):
#     return asyncio.create_task(task)


GenericService = TypeVar("GenericService", bound=Service)



from abc import ABC
from typing import Generic, TypeVar

from web_foundation.kernel.configuration import GenericConfig


class Resource(ABC, Generic[GenericConfig]):
    async def init(self, config: GenericConfig, *args, **kwargs):
        pass

    async def shutdown(self, *args, **kwargs):
        pass


GenericResource = TypeVar("GenericResource", bound=Resource)



from __future__ import annotations

from typing import TypeVar, Generic

from web_foundation import settings
from web_foundation.environment.events.settings import SettingsChange
from web_foundation.environment.metrics.mixin import MetricsMixin
from web_foundation.kernel.abstract.container import GenericDepContainer
from web_foundation.environment.resources.file_repo.repo import FileRepository
from web_foundation.environment.resources.stores.store import AppStore
from web_foundation.kernel import Isolate


class Worker(Isolate, MetricsMixin, Generic[GenericDepContainer]):
    _container: GenericDepContainer
    app_name: str
    _configured: bool
    _store: AppStore | None = None
    _repo: FileRepository | None = None

    def __init__(self, *args, **kwargs):
        self._configured = False

    @property
    def configured(self) -> bool:
        return self._configured

    def _configure(self, *args, **kwargs):
        raise NotImplementedError

    def configure(self, name: str, *args, **kwargs):
        self._configure_isolate(name)
        self._configure(*args, **kwargs)
        self.channel.add_event_listener(SettingsChange, self.on_settings_change)
        self._configured = True

    def post_configure(self):
        self._reg_metrics()
        self._post_configure()

    def _post_configure(self):
        raise NotImplementedError

    async def after_worker_fork(self):
        raise NotImplementedError

    async def _after_worker_fork(self):
        await self._container.init_resources()
        await self.after_worker_fork()

    async def run(self):
        pass

    async def on_settings_change(self, event: SettingsChange):
        setattr(settings, event.name, event.value)

    async def _run(self):
        await self._after_worker_fork()
        await self.run()

    async def close(self, *args, **kwargs):
        await self._container.shutdown_resources()

    @property
    def store(self):
        return self._store

    @store.setter
    def store(self, store: AppStore):
        store.channel = self.channel
        self._store = store

    @property
    def repo(self):
        return self._repo

    @repo.setter
    def repo(self, value: FileRepository):
        self._repo = value

    @property
    def container(self):
        return self._container

    @container.setter
    def container(self, value: GenericDepContainer):
        self._container = value


GenWorker = TypeVar("GenWorker", bound=Worker)



from __future__ import annotations

from abc import ABC
from typing import TypeVar

from web_foundation.kernel.abstract.dependency import Dependency
from web_foundation.utils.helpers import in_obj_classes


class AbstractDependencyContainer(ABC):
    def __init__(self, **kwargs):
        for dep, name in in_obj_classes(self, Dependency):
            dep_init = kwargs.get(name)
            if dep_init:
                dep.from_initialized(dep_init)

    async def init_resources(self):
        pass

    async def shutdown_resources(self):
        pass


GenericDepContainer = TypeVar("GenericDepContainer", bound=AbstractDependencyContainer)



from web_foundation.environment.events.metrics import MetricResponse, MetricRequest
from web_foundation.environment.metrics.basemetric import BaseMetric
from web_foundation.environment.metrics.exporter import MetricExporter
from web_foundation.environment.services.service import Service


class MetricsService(Service):

    async def collect_metrics(self, exporter: MetricExporter):
        metrics_response: MetricResponse = await self.wait_for_response(MetricRequest(exporter))
        return metrics_response.metrics_data

    async def give_metric(self, metric: BaseMetric):
        await self.worker.give_metric(metric)



from __future__ import annotations
import asyncio
from dataclasses import dataclass
from enum import Enum
from typing import Coroutine, Callable, Dict, Type, List, Set, Any
from typing import TypeVar

import loguru
from aioprocessing import AioEvent, AioLock, AioQueue  # type: ignore
from loguru import logger
from web_foundation import settings


class IMessage:
    message_type: str = "__all__"
    index: int
    inner_index: int
    sender: str
    destination: str = "__all__"
    exec_inner: bool = False

    def __init__(self):
        self.index = 0
        self.inner_index = 0
        self.sender = "None"

    def __str__(self):
        return f"{self.__class__.__name__}({self.message_type},{self.index=},{self.inner_index=}, {self.sender=})"


GenericIMessage = TypeVar("GenericIMessage", bound=IMessage, contravariant=True)


@dataclass
class IsolatePipes:
    queue: AioQueue
    event: AioEvent
    lock: AioLock

    async def write(self, msg: GenericIMessage):
        await self.queue.coro_put(msg)

    async def read(self) -> GenericIMessage:
        return await self.queue.coro_get()

    def empty(self):
        return self.queue.empty()


EventListener = Callable[[GenericIMessage], Coroutine]


class MessageWaitState(Enum):
    PENDING = "PENDING"
    DONE = "DONE"


class IMessageWaiter:
    event = asyncio.Event()
    result: Any = None
    call = None

    def __eq__(self, other):
        self.wait_for_index = other.wait_for_index


class IChannel:
    read_timeout = 0.01
    worker_name: str
    consume_pipe: IsolatePipes
    produce_pipe: IsolatePipes

    _listeners: Dict[str, List[EventListener]]
    _once_listeners: Dict[str, EventListener]
    _response_waiters: Dict[int, IMessageWaiter]

    _before_event_middlewares: Dict[str, EventListener]
    _after_event_middlewares: Dict[str, EventListener]
    _inner_index: int

    def __init__(self, isolate_name: str):
        self._inner_index = 0
        self.worker_name = isolate_name
        # self.debug = debug
        self._listeners = {}
        self._once_listeners = {}
        self._response_waiters = {}
        self.produce_pipe = IsolatePipes(AioQueue(), AioEvent(), AioLock())
        self.consume_pipe = IsolatePipes(AioQueue(), AioEvent(), AioLock())
        self._after_event_middlewares = {}
        self._before_event_middlewares = {}

    def __str__(self):
        return f"IChannel(name={self.worker_name})"

    async def produce(self, msg: IMessage):
        self._inner_index += 1
        msg.inner_index = self._inner_index
        msg.sender = self.worker_name
        before_middleware = self._before_event_middlewares.get(msg.message_type)
        if before_middleware:
            await before_middleware(msg)
        if msg.exec_inner:
            listeners = self._listeners.get(msg.message_type)
            for listener in listeners:
                asyncio.create_task(listener(msg))
        await self.produce_pipe.write(msg)

    async def sent_to_consume(self, msg: IMessage):
        await self.consume_pipe.write(msg)

    async def listen_produce(self, callback: Callable[[GenericIMessage], Coroutine]):
        while True:
            r: GenericIMessage = await self.produce_pipe.read()
            if settings.DEBUG:
                logger.info(
                    f"Channel {self.worker_name} send message: {r.message_type}")
            await callback(r)

    async def listen_consume(self):
        while True:
            r: IMessage = await self.consume_pipe.read()
            if settings.DEBUG:
                logger.info(
                    f"Channel {self.worker_name} receive message: {r.message_type}")
            once_callback = self._once_listeners.get(r.message_type)
            if once_callback:
                asyncio.create_task(self._exec_once(r, once_callback))
            after_middleware = self._after_event_middlewares.get(r.message_type)
            if after_middleware:
                asyncio.create_task(after_middleware(r))
            response_waiter = self._response_waiters.get(r.inner_index)
            if response_waiter:
                response_waiter.result = r
            callbacks = self._listeners.get(r.message_type)
            if not callbacks:
                continue
            for callback in callbacks:
                asyncio.create_task(callback(r))

    async def _exec_once(self, msg: GenericIMessage, callback: EventListener):
        await callback(msg)
        self._once_listeners.pop(msg.message_type)

    def add_event_middleware(self, event_type: str, callback: EventListener, assign_to: str = "after"):
        """
        :param callback: callback to call
        :param event_type: "event_message_type"
        :param assign_to: "before, after"
        :return: None
        """
        match assign_to:
            case 'before':
                self._before_event_middlewares.update({event_type: callback})
            case "after":
                self._after_event_middlewares.update({event_type: callback})

    async def produce_for_response(self, msg: GenericIMessage) -> Any:
        task = IMessageWaiter()
        await self.produce(msg)
        self._response_waiters.update({msg.inner_index: task})
        while not task.result:
            await asyncio.sleep(0.01)
        self._response_waiters.pop(msg.inner_index)
        return task.result

    def _add_event_listener(self, event_type: Type[IMessage] | str, callback: EventListener,
                            use_nested_classes: bool = False):
        if use_nested_classes and isinstance(event_type, str):
            raise AttributeError("Can't add_event_listener with use_nested_classes")

        def _add(event_name: str):
            nonlocal self
            nonlocal callback
            if event_name not in self._listeners:
                self._listeners[event_name] = []
            self._listeners[event_name].append(callback)

        if isinstance(event_type, str):
            _add(event_type)
        else:
            def _raise(cls_type):
                if not hasattr(cls_type, "message_type"):
                    raise AttributeError("Can't register listener cause message_type not found")

            if use_nested_classes:
                for cls in event_type.__subclasses__():
                    _raise(cls)
                    _add(cls.message_type)
            else:
                _raise(event_type)
                _add(event_type.message_type)

    def add_event_listener(self, event_type: List[Type[IMessage] | str] | Type[IMessage] | str,
                           callback: EventListener, use_nested_classes: bool = False):
        if isinstance(event_type, list):
            for i in event_type:
                self._add_event_listener(i, callback, use_nested_classes)
        else:
            self._add_event_listener(event_type, callback, use_nested_classes)

    def add_once_listener(self, event_type: str, callback: EventListener):
        self._once_listeners.update({event_type: callback})

    def remove_event_listeners(self, event_type: Type[IMessage] | str):
        if isinstance(event_type, str):
            self._listeners.pop(event_type)
        else:
            self._listeners.pop(event_type.message_type)

    @property
    def listeners(self):
        return self._listeners



import asyncio
import time
from types import SimpleNamespace
from typing import List, Dict

from sanic import Sanic, Request
from sanic.response import BaseHTTPResponse

from web_foundation import settings
from web_foundation.environment.events.metrics import NewMetricEvent
from web_foundation.environment.metrics.basemetric import CounterBaseMetric, HistogramBaseMetric, \
    SummaryBaseMetric
from web_foundation.environment.metrics.exporter import BaseMetric
from web_foundation.kernel import IChannel

try:
    from prometheus_client.metrics import Gauge, Histogram, Counter, MetricWrapperBase
except ImportError:
    settings.PROMETHEUS_METRICS_ENABLE = False


class MetricsMixin:
    _metrics: List[BaseMetric]
    _channel: IChannel

    def _reg_metrics(self):
        pass

    async def give_metric(self, metric: BaseMetric):
        asyncio.create_task(self._channel.produce(NewMetricEvent(metric)))


class ApiMetricsMixin(MetricsMixin):
    _name: str
    app_name: str
    sanic_app: Sanic
    stored_metric: Dict[str, BaseMetric]

    def _reg_metrics(self):
        self.sanic_app.ctx.metrics = SimpleNamespace()

        async def metrics_before(request: Request):
            if not settings.METRICS_ENABLE:
                return
            request.app.ctx.metrics.exec_time = time.time()

        async def metrics_after(request: Request, response: BaseHTTPResponse):
            if not settings.METRICS_ENABLE:
                return
            else:
                metr = HistogramBaseMetric(f"exec_request_time")
                metr.observe(time.time() - request.app.ctx.metrics.exec_time)
                metr.add_label(request=request.path,
                               method=request.method,
                               status=str(response.status))
                await self.give_metric(metr)
                metr = CounterBaseMetric(f"request_path_count", value=1)
                metr.add_label(request=request.path,
                               method=request.method,
                               status=str(response.status))
                await self.give_metric(metr)
                metr = SummaryBaseMetric(f"request_path_latency")
                metr.add_label(request=request.path,
                               method=request.method,
                               status=str(response.status))
                metr.observe(time.time() - request.app.ctx.metrics.exec_time)
                await self.give_metric(metr)

        self.sanic_app.register_middleware(metrics_before)
        self.sanic_app.register_middleware(metrics_after, "response")



from __future__ import annotations
from __future__ import annotations

from importlib.abc import FileLoader
from importlib.util import module_from_spec
from importlib.util import spec_from_loader
from types import ModuleType
from typing import List

import loguru
from loguru import logger

from web_foundation import settings
from web_foundation.environment.events.store import StoreUpdateEvent
from web_foundation.environment.resources.file_repo.repo import FileRepository, AppFileSections, GenFileSections
from web_foundation.environment.resources.stores.store import AppStore

class ApiAddon:
    source: str
    imported: ModuleType | None

    target: str
    name: str
    filename: str
    enabled: bool = False

    before: bool = False
    override: bool = False
    after: bool = False

    def __init__(self, filename: str, source: str, name: str = None):
        self.name = name
        self.filename = filename
        self.source = source
        self.target = ""

    def import_it(self, spec):
        try:
            imported = module_from_spec(spec)
            spec.loader.exec_module(imported)
            self.imported = imported
            if hasattr(self.imported, "before"):
                self.before = True
            if hasattr(self.imported, "after"):
                self.after = True
            if hasattr(self.imported, "override"):
                self.override = True
        except Exception as e:
            self.enabled = False
            loguru.logger.error(f"Can't import plugin \"{self.name}\". Exception: {e}")

    def drop_imported(self):
        self.imported = None
        self.before = False
        self.override = False
        self.after = False

    async def exec_before(self, context, container):
        try:
            if self.before:
                await self.imported.before(context, container)
        except Exception as e:
            loguru.logger.error(f"Can't run plugin \"{self.name}\". Exception: {e}")

    async def exec_after(self, context, container, target_result):
        try:
            if self.after:
                await self.imported.after(context, container, target_result)
        except Exception as e:
            loguru.logger.error(f"Can't run plugin \"{self.name}\". Exception: {e}")

    async def exec_override(self, context, container):
        try:
            if self.override:
                return await self.imported.override(context, container)
        except Exception as e:
            loguru.logger.error(f"Can't run plugin \"{self.name}\". Exception: {e}")

    def __repr__(self):
        return f"ApiMiddleware({self.name} on {self.target})"

    def __eq__(self, other: ApiAddon):
        return self.name == other.name


class SourceCodeLoader(FileLoader):
    def __init__(self, fullname: str, source):
        super().__init__(fullname, source)
        self.path = source

    def get_source(self, fullname: str) -> str | bytes:
        return self.path


class AddonsLoader:
    repo: FileRepository = None
    store: AppStore = None
    sections: GenFileSections

    def __init__(self, sections: GenFileSections = AppFileSections):
        self.sections = sections

    async def _import_middleware(self, pl: ApiAddon):
        middlewares = await self.store.get_item("middleware") if await self.store.get_item("middleware") else []
        try:
            spec = spec_from_loader(pl.name, loader=SourceCodeLoader(pl.name, pl.source))
            pl.import_it(spec)
            if pl not in middlewares:
                middlewares.append(pl)
            await self.store.set_item("middleware", middlewares, send_event=False)
        except Exception as e:
            if settings.DEBUG:
                logger.debug(f"Can't load plugin {pl.name}, cause {str(e)}")

    async def import_middleware(self):
        middlewares = await self.store.get_item("middleware") if await self.store.get_item("middleware") else []
        for pl in middlewares:
            await self._import_middleware(pl)

    async def configure_middleware(self, plugin: ApiAddon, *args, **kwargs) -> ApiAddon:
        """Set middleware name, target and enabled"""
        raise NotImplementedError

    async def _discover_middleware(self, filename) -> ApiAddon | None:
        async with (await self.repo.take(filename, self.sections.PLUGINS)) as file_:
            pl = ApiAddon(filename, await file_.read())
            pl = await self.configure_middleware(pl)
            if pl.name is None:
                raise AttributeError("ApiMiddleware name must be specified")

            middlewares = await self.store.get_item("middleware") if await self.store.get_item("middleware") else []
            if middlewares and pl in middlewares:
                middlewares.remove(pl)
            middlewares.append(pl)
            await self.store.set_item("middleware", middlewares, obj=True)
            return pl

    async def discover_middleware(self):
        for filename in await self.repo.open_section(self.sections.PLUGINS):
            await self._discover_middleware(filename)

    async def delete_middleware(self, filename: str):
        mdws = await self.store.get_item("middleware")
        new_mdws: List[ApiAddon] = []
        for mdw in mdws:
            if mdw.filename != filename:
                new_mdws.append(mdw)
        await self.store.set_item("middleware", new_mdws, obj=True)
        await self.import_middleware()

    async def add_new_middleware(self, filename: str):
        pl = await self._discover_middleware(filename)
        if pl:
            await self._import_middleware(pl)
            if settings.DEBUG:
                logger.debug(f"New plugin added {pl}")

    async def find_middleware_by_target(self, target: str) -> List[ApiAddon]:
        middlewares = await self.store.get_item("middleware")
        if not middlewares:
            return []
        else:
            cell_plugins: List[ApiAddon] = []
            for pl in middlewares:
                if pl.target == target and pl.enabled:
                    cell_plugins.append(pl)
            return cell_plugins

    async def on_store_update(self, event: StoreUpdateEvent):
        if event.key == "middleware":
            await self.store.set_item("middleware", event.value, send_event=False)
            await self.import_middleware()

    def reg_channel_middleware(self):
        self.store.channel.add_event_middleware(StoreUpdateEvent.message_type, self.on_store_update)
        self.store.channel.add_event_middleware(StoreUpdateEvent.message_type, self.on_mdw_store_set,
                                                assign_to="before")

    async def on_mdw_store_set(self, event: StoreUpdateEvent):
        if event.key == "middleware":
            for mdw in event.value:
                mdw.imported = None

    async def first_init(self):
        await self.discover_middleware()
        await self.import_middleware()
